@@ -157,7 +157,7 @@ class AppMenuButtonRightClickMenu extends Applet.AppletPopupMenu {
         this.signals.connect(item, 'activate', () => this.state.trigger('configureApplet'));
         subMenu.menu.addMenuItem(item);
 
-        item = createMenuItem({label: _('Remove') + " 'Grouped window list'", icon: 'edit-delete'});
+        item = createMenuItem({label: _("Remove '%s'").format(_("Grouped window list")), icon: 'edit-delete'});
         this.signals.connect(item, 'activate', () => {
             AppletManager._removeAppletFromPanel(this.state.uuid, this.state.instance_id);
         });
@@ -278,10 +278,10 @@ class AppMenuButtonRightClickMenu extends Applet.AppletPopupMenu {
                 let label, icon;
                 if (this.groupState.isFavoriteApp) {
                     label = _('Unpin from Panel');
-                    icon = 'gwl-unpin';
+                    icon = 'unpin';
                 } else {
                     label = _('Pin to Panel');
-                    icon = 'gwl-pin';
+                    icon = 'pin';
                 }
                 this.pinToggleItem = createMenuItem({label, icon});
                 this.signals.connect(this.pinToggleItem, 'activate', (...args) => this.toggleFavorite(...args));
@@ -460,15 +460,34 @@ class AppMenuButtonRightClickMenu extends Applet.AppletPopupMenu {
 
     destroy() {
         this.signals.disconnectAllSignals();
-        Applet.AppletPopupMenu.prototype.destroy.call(this);
+        super.destroy();
         unref(this, RESERVE_KEYS);
     }
 }
 
 class HoverMenuController extends PopupMenu.PopupMenuManager {
+    constructor(actor, groupState) {
+        super({actor}, false); // owner, shouldGrab
+        this.groupState = groupState;
+        this.connectId = this.groupState.connect({
+            thumbnailMenuEntered: ({thumbnailMenuEntered}) => {
+                this.shouldGrab = thumbnailMenuEntered;
+                this._onMenuOpenState(this._menus[0], this._menus[0].isOpen);
+                this.groupState.trigger('checkFocusStyle');
+                if (!this.grabbed) return;
+                if (!thumbnailMenuEntered) this._ungrab();
+            }
+        });
+    }
+
     _onEventCapture() {
         return false;
     }
+
+    destroy() {
+        this.groupState.disconnect(this.connectId);
+        super.destroy();
+    }
 }
 
 class WindowThumbnail {
@@ -563,7 +582,8 @@ class WindowThumbnail {
         this.signals.connect(this.button, 'button-release-event', (...args) => this.onCloseButtonRelease(...args));
         this.signals.connect(this.actor, 'button-release-event', (...args) => this.connectToWindow(...args));
 
-        setTimeout(() => this.handleFavorite(), 0);
+
+        this.handleFavorite();
         // Update focused style
         this.onFocusWindowChange();
     }
@@ -665,35 +685,30 @@ class WindowThumbnail {
         this.stopClick = false;
     }
 
-    getThumbnail() {
+    getThumbnail(thumbnailWidth, thumbnailHeight) {
         if (this.groupState.verticalThumbs || !this.state.settings.showThumbs) {
-            this.thumbnailActor.height = 0;
+            this.thumbnailActor.hide();
             return null;
+        } else if (this.thumbnailActor.realized) {
+            this.thumbnailActor.show();
         }
         // Create our own thumbnail if it doesn't exist
-        let isUpdate = false;
         if (this.metaWindowActor) {
-            isUpdate = true;
             this.signals.disconnect('size-changed', this.metaWindowActor);
+        } else {
+            this.metaWindowActor = this.metaWindow.get_compositor_private();
         }
-        this.metaWindowActor = this.metaWindow.get_compositor_private();
         if (this.metaWindowActor) {
-            let windowTexture = this.metaWindowActor.get_texture();
-            let [width, height] = windowTexture.get_size();
             this.signals.connect(this.metaWindowActor, 'size-changed', () => this.refreshThumbnail());
 
-            if (this.groupState.isFavoriteApp) {
-                this.signals.connect(this.metaWindowActor, 'destroy', () => {
-                    if (this.willUnmount || !this.groupState.trigger) return;
-                    this.groupState.trigger('removeThumbnailFromMenu', this.metaWindow);
-                    this.metaWindowActor = null;
-                });
-            }
-
-            let scale = Math.min(1.0, this.thumbnailWidth / width, this.thumbnailHeight / height) * global.ui_scale;
+            let windowTexture = this.metaWindowActor.get_texture();
+            let [width, height] = windowTexture.get_size();
+            let scale = Math.min(1.0, thumbnailWidth / width, thumbnailHeight / height) * global.ui_scale;
             width = Math.round(width * scale);
             height = Math.round(height * scale);
-            if (isUpdate) {
+            if (this.thumbnailActor.child) {
+                this.thumbnailActor.height = height;
+                this.thumbnailActor.width = width;
                 this.thumbnailActor.child.source = windowTexture;
                 this.thumbnailActor.child.width = width;
                 this.thumbnailActor.child.height = height;
@@ -720,22 +735,29 @@ class WindowThumbnail {
         }
 
         let monitor = this.state.trigger('getPanelMonitor');
+        if (!monitor) return;
 
         if (!this.thumbnailActor || this.thumbnailActor.is_finalized()) return;
 
-        let divider = 80;
-        let offset = 16;
+        let divider = 80 * global.ui_scale;
+        let {thumbSize} = this.state.settings;
 
-        this.thumbnailWidth = Math.floor((monitor.width / divider) * this.state.settings.thumbSize) + offset;
-        this.thumbnailHeight = Math.floor((monitor.height / divider) * this.state.settings.thumbSize) + offset;
+        if (monitor.height / global.ui_scale <= 1024) {
+            thumbSize += 6;
+        } else if (monitor.height / global.ui_scale <= 1200) {
+            thumbSize += 3;
+        }
+
+        let thumbnailWidth = Math.floor((monitor.width / divider) * thumbSize);
+        let thumbnailHeight = Math.floor((monitor.height / divider) * thumbSize);
 
         let monitorSize, thumbnailSize;
         if (!this.state.isHorizontal) {
             monitorSize = monitor.height;
-            thumbnailSize = this.thumbnailHeight;
+            thumbnailSize = thumbnailHeight;
         } else {
             monitorSize = monitor.width;
-            thumbnailSize = this.thumbnailWidth;
+            thumbnailSize = thumbnailWidth;
         }
 
         let padding = this.thumbnailActor.style_length('padding');
@@ -743,10 +765,10 @@ class WindowThumbnail {
 
         let i = 0;
         while (((thumbnailSize + this.thumbnailPadding + padding + margin) * this.groupState.windowCount > monitorSize)
-            && this.thumbnailWidth > 64
-            && this.thumbnailHeight > 64) {
-            this.thumbnailWidth -= 1;
-            this.thumbnailHeight -= 1;
+            && thumbnailWidth > 64
+            && thumbnailHeight > 64) {
+            thumbnailWidth -= 1;
+            thumbnailHeight -= 1;
             thumbnailSize -= 1;
             i++;
             // Bail after 200 iterations
@@ -762,17 +784,17 @@ class WindowThumbnail {
         this.groupState.set({verticalThumbs});
         if (verticalThumbs !== currentVerticalThumbsState) return;
 
-        let scaledWidth = this.thumbnailWidth * global.ui_scale;
+        let scaledWidth = thumbnailWidth * global.ui_scale;
         this.thumbnailActor.width = scaledWidth;
-        this.container.style = `width: ${Math.floor(this.thumbnailWidth - 16)}px;`;
+        this.container.style = `width: ${Math.floor(thumbnailWidth - 16)}px;`;
         if (this.groupState.verticalThumbs || (this.state.settings.verticalThumbs && this.state.settings.showThumbs)) {
-            this.thumbnailActor.height = this.thumbnailHeight;
+            this.thumbnailActor.height = thumbnailHeight;
         } else if (this.state.settings.verticalThumbs) {
             this.thumbnailActor.height = 0;
         }
 
         this.labelContainer.child.text = this.metaWindow.title || '';
-        this.getThumbnail();
+        this.getThumbnail(thumbnailWidth, thumbnailHeight);
     }
 
     hoverPeek(opacity) {
@@ -839,9 +861,16 @@ class AppThumbnailHoverMenu extends PopupMenu.PopupMenu {
         this.connectId = this.groupState.connect({
             hoverMenuClose: () => {
                 this.shouldClose = true;
+                this.groupState.set({thumbnailMenuEntered: false});
                 this.close();
             },
-            addThumbnailToMenu: (win) => this.addThumbnail(win),
+            addThumbnailToMenu: (win) => {
+                if (this.isOpen) {
+                    setTimeout(() => this.addThumbnail(win), 0);
+                    return;
+                }
+                this.queuedWindows.push(win);
+            },
             removeThumbnailFromMenu: (win) => {
                 let index = findIndex(this.appThumbnails, (item) => item.metaWindow === win);
                 if (index > -1) {
@@ -849,16 +878,20 @@ class AppThumbnailHoverMenu extends PopupMenu.PopupMenu {
                     this.appThumbnails[index] = undefined;
                     this.appThumbnails.splice(index, 1);
                 }
+                index = this.queuedWindows.indexOf(win);
+                if (index > -1) this.queuedWindows.splice(index, 1);
             },
             verticalThumbs: () => {
                 // Preserve the menu's open state after refreshing
                 let {isOpen} = this;
-                this.setVerticalSetting()
+                this.setVerticalSetting();
                 if (isOpen) this.open(true);
-            }
+            },
         });
 
         this.appThumbnails = [];
+        this.queuedWindows = [];
+        this.fullyRefreshThumbnails();
     }
 
     onButtonPress() {
@@ -869,30 +902,45 @@ class AppThumbnailHoverMenu extends PopupMenu.PopupMenu {
         setTimeout(() => this.close(), this.state.settings.thumbTimeout);
     }
 
-    onMenuEnter() {
+    onMenuEnter(actor) {
         if (this.state.panelEditMode ||
             (!this.isOpen && this.state.settings.onClickThumbs) ||
             this.state.menuOpen) {
             return false;
         }
+
         this.shouldClose = false;
 
         let timeout;
-
         if (this.state.thumbnailMenuOpen) {
             timeout = 50;
         } else {
             timeout = this.state.settings.thumbTimeout;
         }
 
+        if (this.queuedWindows.length > 0) {
+            each(this.queuedWindows, (win) => this.addThumbnail(win));
+            this.queuedWindows = [];
+        }
+
+        if (actor != null) {
+            this.groupState.set({thumbnailMenuEntered: this.isOpen});
+        }
+
         setTimeout(() => this.open(), timeout);
     }
 
-    onMenuLeave() {
+    onMenuLeave(actor) {
         if (this.state.menuOpen || this.state.panelEditMode) {
             return false;
         }
+
         this.shouldClose = true;
+
+        if (actor != null) {
+            this.groupState.set({thumbnailMenuEntered: false});
+        }
+
         setTimeout(() => this.close(), 50);
     }
 
@@ -917,7 +965,7 @@ class AppThumbnailHoverMenu extends PopupMenu.PopupMenu {
             this.groupState.tooltip.show();
         } else {
             this.state.set({thumbnailMenuOpen: true});
-            PopupMenu.PopupMenu.prototype.open.call(this, this.state.settings.animateThumbs);
+            super.open(this.state.settings.animateThumbs);
         }
     }
 
@@ -928,13 +976,14 @@ class AppThumbnailHoverMenu extends PopupMenu.PopupMenu {
             || !this.groupState.tooltip) {
             return;
         }
-        if (!this.groupState.metaWindows || this.groupState.metaWindows.length === 0) {
+        if ((!this.groupState.metaWindows || this.groupState.metaWindows.length === 0)
+            && !this.groupState.tooltip._tooltip.is_finalized()) {
             this.groupState.tooltip.set_text('');
             this.groupState.tooltip.hide();
         }
         if (this.isOpen) {
             this.state.set({thumbnailMenuOpen: false});
-            PopupMenu.PopupMenu.prototype.close.call(this, this.state.settings.animateThumbs);
+            if (!this.actor.is_finalized()) super.close(this.state.settings.animateThumbs);
         }
         for (let i = 0; i < this.appThumbnails.length; i++) {
             this.appThumbnails[i].destroyOverlayPreview();
@@ -942,6 +991,9 @@ class AppThumbnailHoverMenu extends PopupMenu.PopupMenu {
     }
 
     onKeyPress(actor, e) {
+        let {orientation} = this.state;
+        let {vertical} = this.box;
+
         let symbol = e.get_key_symbol();
         let i = findIndex(this.appThumbnails, (item) => item.entered === true);
         let entered = i > -1;
@@ -955,19 +1007,25 @@ class AppThumbnailHoverMenu extends PopupMenu.PopupMenu {
         }
         let args;
         let closeArg;
-        if (this.state.orientation === St.Side.TOP) {
+        if (orientation === St.Side.TOP) {
             closeArg = Clutter.KEY_Up;
             args = [Clutter.KEY_Left, Clutter.KEY_Right];
-        } else if (this.state.orientation === St.Side.BOTTOM) {
+        } else if (orientation === St.Side.BOTTOM) {
             closeArg = Clutter.KEY_Down;
             args = [Clutter.KEY_Right, Clutter.KEY_Left];
-        } else if (this.state.orientation === St.Side.LEFT) {
+        } else if (orientation === St.Side.LEFT) {
             closeArg = Clutter.KEY_Left;
             args = [Clutter.KEY_Up, Clutter.KEY_Down];
-        } else if (this.state.orientation === St.Side.RIGHT) {
+        } else if (orientation === St.Side.RIGHT) {
             closeArg = Clutter.KEY_Right;
             args = [Clutter.KEY_Down, Clutter.KEY_Up];
         }
+
+        // Panel is oriented horizontally, but the menu is vertical
+        if (vertical && (orientation === St.Side.TOP || orientation === St.Side.BOTTOM)) {
+            args = [Clutter.KEY_Down, Clutter.KEY_Up];
+        }
+
         let index;
         if (symbol === args[0]) {
             if (!entered) {
@@ -1006,7 +1064,7 @@ class AppThumbnailHoverMenu extends PopupMenu.PopupMenu {
             this.destroyThumbnails();
         }
         this.addWindowThumbnails(this.groupState.metaWindows);
-        this.setStyleOptions(false);
+        this.setStyleOptions();
     }
 
     destroyThumbnails() {
@@ -1068,12 +1126,13 @@ class AppThumbnailHoverMenu extends PopupMenu.PopupMenu {
         }
     }
 
-    setStyleOptions(skipThumbnailIconResize) {
+    setStyleOptions() {
         if (this.willUnmount || !this.box) return;
 
         // The styling cannot be set correctly unless the menu is closed. Fortunately this
         // can be closed and reopened too quickly for the user to notice.
-        if (this.isOpen) this.close(true);
+        let {isOpen} = this;
+        if (isOpen) this.close(true);
 
         this.box.show();
         this.box.style = null;
@@ -1087,9 +1146,7 @@ class AppThumbnailHoverMenu extends PopupMenu.PopupMenu {
         let boxPadding = padding && padding > 0 ? padding : 3;
         this.box.style = `padding: ${boxPadding}px;`;
 
-        if (skipThumbnailIconResize) return;
-
-        if (this.isOpen) this.open();
+        if (isOpen) this.open(true);
     }
 
     setVerticalSetting() {
@@ -1098,15 +1155,27 @@ class AppThumbnailHoverMenu extends PopupMenu.PopupMenu {
         } else {
             this.box.vertical = true;
         }
-        this.fullyRefreshThumbnails();
+
+        // Do a full refresh if thumbnails don't exist - this happens when the thumbnail menu
+        // initializes vertically from lack of calculated space, or thumbnails are disabled.
+        if (!this.appThumbnails[0] || !this.appThumbnails[0].thumbnailActor.child) {
+            this.fullyRefreshThumbnails();
+        } else {
+            this.updateThumbnailSize();
+        }
     }
 
     updateThumbnailSize() {
         for (let i = 0; i < this.appThumbnails.length; i++) {
             if (this.appThumbnails[i]) {
                 this.appThumbnails[i].refreshThumbnail();
+                // Make sure Clutter updates, otherwise setStyleOptions is called afterwards,
+                // and will be calculating styles from old actor values because it closes the menu
+                // to avoid incorrect padding values.
+                this.appThumbnails[i].thumbnailActor.realize();
             }
         }
+        this.setStyleOptions();
     }
 
     destroy() {
@@ -1126,7 +1195,7 @@ class AppThumbnailHoverMenu extends PopupMenu.PopupMenu {
             }
         }
         this.removeAll();
-        PopupMenu.PopupMenu.prototype.destroy.call(this);
+        super.destroy();
         this.groupState.disconnect(this.connectId);
         unref(this, RESERVE_KEYS);
     }
